function parseSpeechConfig(xml) {
    var functions = {};
    var i;
    var root = parseNode(xml.documentElement, functions);
    functions["main"] = root.children;
    return functions;
}
    
function parseNode(n, functions) {
    function makeSoundPlayer(f) {
        return function(context) {
            context.playSound(f);
        };
    }
    function makeBitSetter(bits) {
        return function(context) {
            context.setBitFlags(bits);
        };
    }
    function makeStateSetter(s) {
        return function(context) {
            context.setState(s);
        };
    }
    function makeFlagSetter(f, clear) {
        if (clear) {
            return function(context) {
                context.setFlag(f);
            };
        }
        else {
            return function(context) {
                context.clearFlag(f);
            };
        }    
    }
    function attVal(att) {
        if (att) return att.nodeValue;
    }
    var els;
    var i;
    var child;
    
    if (!(n instanceof Element)) {
        return;
    }
    if (n.nodeName === "text") {
        // don't care
        return;
    }
    if (n.nodeName === "speech") {
        // skip through speech element
        for (i = 0; n.childNodes && i < n.childNodes.length; i++) {
            child = parseNode(n.childNodes[i], functions);
            if (child) {
                return child;
            }
        }
    }
    if (n.nodeName === "link") {
        els = [];
       
        for (i = 0; n.childNodes && i < n.childNodes.length; i++) {
            child = parseNode(n.childNodes[i], functions);
            if (child) {
                els.push(child);
            }
        }
        functions[attVal(n.attributes.id)] = els;
        return;
    }
    var obj = { type: "normal"}
    obj.logic = attVal(n.attributes.logic);
    
    if (n.nodeName === "dummy") {
        obj.type = "dummy";
        obj.retFalse = true;
    }
    else if (n.nodeName === "ref") {
        obj.type = "call";
        obj.ref = attVal(n.attributes.ref);
    }
    else if (n.nodeName === "state") {
        obj.state = attVal(n.attributes.val);
    }
    else if (n.nodeName === "flags") {
        obj.flag = attVal(n.attributes.val);
        obj.logic = attVal(n.attributes.logic);
    }
    else if (n.nodeName === "maneuver") {
        obj.maneuver = attVal(n.attributes.val);
    }
    else if (n.nodeName === "turn") {
        obj.turn = attVal(n.attributes.val);
    }
    else if (n.nodeName === "roundaboutexit") {
        obj.roundaboutexit = attVal(n.attributes.val);
    }
    else if (n.nodeName === "maneuver_at_endofstreet") {
        obj.maneuver_at_endofstreet = true;
    }
    else if (n.nodeName === "nextmaneuver") {
        obj.nextmaneuver = attVal(n.attributes.val);
    }
    else if (n.nodeName === "nextturn") {
        obj.nextturn = attVal(n.attributes.val);
    }
    else if (n.nodeName === "nextroundaboutexit") {
        obj.nextroundaboutexit = attVal(n.attributes.val);
    }
    else if (n.nodeName === "distancetomaneuver") {
        obj.distMinTo = attVal(n.attributes.min);
        obj.distMaxTo = attVal(n.attributes.max);
    }
    else if (n.nodeName === "nextmaneuverdistance") {
        obj.nextmaneuverdistanceMax = attVal(n.attributes.max);
    }
    else if (n.nodeName === "distanceposition") {
        /* FIXME WRONG 
           distanceposition is NOT the same as distancetomaneuver as in some places 
           in config.xml distanceposition is compared to a small value and then distancetomaneuver 
           is immediately compared to a large value
	*/
        obj.checkBitsNotSet = attVal(n.attributes.testflagn);
        obj.distMinTo = attVal(n.attributes.minto);
        obj.distMinFrom = attVal(n.attributes.minfrom);
        obj.distMaxTo = attVal(n.attributes.maxto);
        obj.distMaxFrom = attVal(n.attributes.maxfrom);
        if (n.attributes.setflagp) {
            obj.action = makeBitSetter(attVal(n.attributes.setflagp));
        }
    }
    else if (n.nodeName === "timeposition") {
        obj.checkBitsNotSet = attVal(n.attributes.testflagn);
        obj.timeMinTo = attVal(n.attributes.minto);
        obj.timeMinFrom = attVal(n.attributes.minfrom);
        obj.timeMaxTo = attVal(n.attributes.maxto);
        obj.timeMaxFrom = attVal(n.attributes.maxfrom);
        if (n.attributes.setflagp) {
            obj.action = makeBitSetter(attVal(n.attributes.setflagp));
        }
    }
    else if (n.nodeName === "setstate") {
        obj.action = makeStateSetter(attVal(n.attributes.setflagp));
        obj.retFalse = true;
    }
    else if (n.nodeName === "setflags") {
        if (n.attributes.setflagp) {
            obj.action = makeFlagSetter(attVal(n.attributes.setflagp));
        }
        else {
            obj.action = makeFlagSetter(attVal(n.attributes.setflagn), true);
        }
        obj.retFalse = true;
    }
    else if (n.nodeName === "soundfile") {
        obj.action = makeSoundPlayer(attVal(n.attributes.val));
        obj.retFalse = true;
    }
    
    
    els = [];
       
    for (i = 0; n.childNodes && i < n.childNodes.length; i++) {
        child = parseNode(n.childNodes[i], functions);
        if (child) {
            els.push(child);
        }
    }
    obj.children = els;
    return obj;
}

function evaluate(context, elements) {
    var i = 0;
    
    for (; i < elements.length; i++) {
        if (evaluateElement(context, elements[i])) {
            return true;
        }
    }
    return false;
}

function evaluateElement(context, el) {
    var callResult;
    
    if (el.type === "call") {
         if (evaluate(context, context.functions[el.ref])) {
            if (el.children) {
                evaluate(context, el.children);
            }
            return true;
         }
         return false;
    }
    else {
        if ((el.state && !context.isStateSet(el.state))
        || (el.checkBitsNotSet && context.bitsSet(el.checkBitsNotSet))
        || (el.flag && !context.isFlagSet(el.flag))
        || (el.maneuver && context.getManeuver() !== el.maneuver) 
        || (el.turn && context.getTurn() !== el.turn)
        || (el.roundaboutexit && context.getRoundaboutExit() !== el.roundaboutexit)
        || (el.maneuver_at_endofstreet && !context.isManeuverAtEndOfStreet())
        || (el.nextmaneuver && context.getNextManeuver() !== el.nextmaneuver)
        || (el.nextturn && context.getNextTurn() !== el.nextturn)
        || (el.nextroundaboutexit && context.getNextRoundaboutExit() !== el.nextroundaboutexit)
        || (el.nextmaneuverdistanceMax && context.getDistanceToNext() > el.nextmaneuverdistanceMax)
        || (el.distMinTo && context.getDistanceTo() < el.distMinTo)
        || (el.distMaxTo && context.getDistanceTo() > el.distMaxTo)
        || (el.distMinFrom && context.getDistanceFrom() < el.distMinFrom) /* is 0 a special value? It is in config.xml in a way that doesn't seem to make sense */
        || (el.distMaxFrom && context.getDistanceFrom() > el.distMaxFrom)
        || (el.timeMinTo && context.getTimeTo() < el.timeMinTo)
        || (el.timeMaxTo && context.getTimeTo() > el.timeMaxTo)
        || (el.timeMinFrom && context.getTimeFrom() < el.timeMinFrom)
        || (el.timeMaxFrom && context.getTimeFrom() > el.timeMaxFrom)) {
            return false;
        }
        
        if (el.action) {
            el.action(context);
        }
        
        
        callResult = evaluate(context, el.children);
        // logic attribute some controls && and || sorta behaviour. Not sure what 1 and 2 mean yet
        if (el.retFalse) { 
            return false;
        }
        if (el.logic == 2) {
            return callResult;
        }
        return true;
    }

}

var eg = {
    state: { "WITH_ROUTE": true, "ON_TARGET": true, "ON_HIGHWAY": true },
    flags: { "KM": true },
    bits: 0,
    maneuver: "END",
    turn: "LIGHTRIGHT",
    roundaboutexit: 6,
    maneuver_at_endofstreet: false,
    nextmaneuver: "JUNCTION",
    turn: "LIGHTRIGHT",
    roundaboutexit: null,
    distanceto: 1000,
    timeto: 3,
    distancefrom: 2500,
    timefrom: 35,
    distancetonext: 2006
}

function makeContext(obj) {
    return {
        /* getters */
        isStateSet: function(state) { return obj.state[state];},
        isFlagSet: function(flag) { return obj.flags[flag];},
        bitsSet: function(bits) { return bits & obj.bits;},
        getManeuver: function() { return obj.maneuver; },
        getTurn: function() { return obj.turn; },
        getRoundaboutExit: function() { return obj.roundaboutexit; },
        isManeuverAtEndOfStreet: function() { return obj.maneuver_at_endofstreet; },
        getNextManeuver: function() { return obj.nextmaneuver; },
        getNextTurn: function() { return obj.nextturn; },
        getNextRoundaboutExit: function() { return obj.nextroundaboutexit; },
        getDistanceTo: function() { return obj.distanceto; },
        getDistanceFrom: function() { return obj.distancefrom; },
        getTimeTo: function() { return obj.timeto; },
        getTimeFrom: function() { return obj.timefrom; },
        getDistanceToNext: function() { return obj.distancetonext; },
        
        /* setters */
        setBitFlags: function(bits) {
            obj.bits = obj.bits | bits;
        },
        setState: function(state) {
            /* dunno what this is for */
        },
        setFlag: function(flag) {
            obj.flags[flag] = true;
        },
        clearFlag: function(flag) {
            delete obj.flags[flag];
        },
        playSound: function(file) {
            // TODO
            console.info(file);
        }
    };
};